const cherow = require('cherow')
const esquery = require('esquery')
const astring = require('astring')
const estraverse = require('estraverse')
const template = require('estemplate')
const comparify = require('comparify')
const toAST = require('to-ast')
const prettier = require('prettier')
const sourcemap = require('source-map')
const replace = require('./src/replace')

class AbstractSyntaxTree {
  constructor (source, options) {
    options = options || {}
    if (typeof source === 'string') {
      this.source = source
      this.ast = this.constructor.parse(source, {
        loc: true,
        jsx: options.jsx
      })
    } else {
      this.ast = source
    }
  }

  query (node, selector) {
    return esquery(node, selector)
  }

  find (selector, options) {
    return this.query(this.ast, selector, options)
  }

  each (selector, callback) {
    return this.find(selector).forEach(callback)
  }

  first (selector) {
    return this.find(selector)[0]
  }

  last (selector) {
    var nodes = this.find(selector)
    return nodes[nodes.length - 1]
  }

  count (selector) {
    return this.find(selector).length
  }

  has (selector) {
    return this.count(selector) > 0
  }

  is (node, expected) {
    return comparify(node, expected)
  }

  remove (target, options) {
    options = options || {}
    if (typeof target === 'string') {
      return this.removeBySelector(target, options)
    }
    this.removeByNode(target, options)
  }

  removeBySelector (target, options) {
    var nodes = this.find(target)
    // this could be improved by traversing once and
    // comparing the current node to the found nodes
    // one by one while making the array of nodes smaller too
    nodes.forEach(node => this.removeByNode(node, options))
  }

  removeByNode (node, options) {
    var count = 0
    estraverse.replace(this.ast, {
      enter: function (current, parent) {
        if (options.first && count === 1) {
          return this.break()
        }
        if (comparify(current, node)) {
          count += 1
          return this.remove()
        }
      },
      leave: function (current, parent) {
        if (current.expression === null ||
                  (current.type === 'VariableDeclaration' && current.declarations.length === 0)) {
          return this.remove()
        }
      }
    })
  }

  walk (callback) {
    return estraverse.traverse(this.ast, { enter: callback })
  }

  traverse (options) {
    return estraverse.traverse(this.ast, options)
  }

  replace (options) {
    return replace(this.ast, options)
  }

  prepend (node) {
    this.ast.body.unshift(node)
  }

  append (node) {
    this.ast.body.push(node)
  }

  body () {
    return this.ast.body
  }

  wrap (callback) {
    this.ast.body = callback(this.ast.body)
  }

  unwrap () {
    let block = this.first('BlockStatement')
    this.ast.body = block.body
  }

  template (source, options) {
    options = options || {}
    if (typeof source === 'string') {
      return template(source, options).body
    }
    return toAST(source, options)
  }

  beautify (source, options) {
    if (typeof options === 'boolean') { options = {} }
    options = { parser: 'babylon', ...options }
    return prettier.format(source, options)
  }

  mark () {
    let cid = 1
    this.walk(node => {
      node.cid = cid
      cid += 1
    })
  }

  minify (ast) {
    return ast
  }

  toSource (options) {
    options = options || {}

    if (options.minify) {
      this.ast = this.minify(this.ast)
    }

    var source = this.constructor.generate(this.ast)

    if (options.beautify) {
      source = this.beautify(source, options.beautify)
    }

    var map
    if (options.sourceMap) {
      map = this.toSourceMap(options)
    }

    this.source = source

    if (map) { return { map, source } }
    return source
  }

  toSourceMap (options) {
    const map = new sourcemap.SourceMapGenerator({
      file: options.sourceFile || 'UNKNOWN'
    })
    this.constructor.generate(this.ast, {
      sourceMap: map
    })
    return map.toString()
  }

  toString (options) {
    return this.toSource(options)
  }

  static generate (tree, options) {
    return astring.generate(tree, options)
  }

  static parse (source, options) {
    return cherow.parseModule(source, options)
  }

  static walk (node, callback) {
    return estraverse.traverse(node, { enter: callback })
  }

  static replace (node, callback) {
    return replace(node, { enter: callback })
  }
}

module.exports = AbstractSyntaxTree
